{ **********************************************************************

    Copyright (c) 2013 by Bruno Krayenbuhl

    FPC (Free pascal) implementation of hash BASED (but not identical) on
        Fowler / Noll / Vo (FNV) hash.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 ****************************************************************************}

{ **********************************************************************

    According to Landon Curt Noll web site :
    "http://www.isthe.com/chongo/tech/comp/fnv/index.html"
    and Wikipedia :
    "http://en.wikipedia.org/wiki/Fowler_Noll_Vo_hash"
    FNV hash is in the public domain.

 **********************************************************************}

{ Abstract:
    HashFNV32Buf  is a function that produces hash values for buffer's of data.
    PackHash32To8 is a function that produces a byte combining a 32 bit hash
                  value using xor.
}
{ History :
  2014.08.27 try to generate same FNV for BIG and LITTLE ENDIAN processors
             not tested <- remove me if works }

unit uFnvHash32buf;

interface
var
  FNVLengthShouldBeOdd:boolean=False;    // Debug : passed aLength should be odd

// Function params :
//    aPointer : pointer to data buffer
//    aLength  : length in bytes of data pointed by aPointer to be hashed
function HashFNV32Buf(const aPointer:pointer; const aLength:DWord):DWord;
function PackHash32To8(aDWord: DWord): byte;

const
  CFNV1_32BS:DWord = 4294967279; // 4294967279 prime to try to reseed the sticky
                                 // condition (#BK) returned if aLength = 0

implementation

const
  CFNV1_32P:DWord  = 16777619;   // FNV-1 32 bit prime Fowler / Noll / Vo (FNV) Hash
  CFNV1_32B:DWord  = 4294967291; // 4294967291 Largest prime fitting in a DWord.
                           // On an i386 it seems to have a better spread than
                           // the 32 bit offset_basis = 2166136261 as per FNV.
                           // 4294967291 has the high bit set.

// !!! Warning !!! to test : equality of results for BIG_ENDIAN and
//                           LITTLE_ENDIAN processors
// !!! Warning !!! to test : how does it behave when aPointer is not a Word
//                           aligned pointer (ARM ???)
// This function is NOT INTENDED to be a cryptographic hash but to be
// used to find and differentiate buffers contents faster.
// Example :
// using an Adelson-Velskii / Landis tree (AVL) containing data records in the
// form (HashFNV32Buf(SomeData, Length(Buffer)) : DWord; SomeData : buffer)
//        and sorted on (HashFNV32Buf, sort_order(SomeData)).
// If the tree is large, it will be faster (how much ?) to find hits or
// detect inexistent values when first comparing the DWORD HashFNV32Buf value.
// Same can be said for ordered records searched with a binary search and maybe
// some other search on ordered tables and arrays.
function HashFNV32Buf(const aPointer: pointer; const aLength: DWord): DWord;
var
  lLength:integer;
  lDWord:DWord;
  lpByte:pointer;
  lPointer:pointer;
begin
  {
  if FNVLengthShouldBeOdd then
    Assert((aLength and 1)<>0);
  }
  if aLength=0 then begin           // Return sticky resetter value
    Result:=CFNV1_32BS;
    exit;
  end;
  lLength:=aLength;
  lPointer:=aPointer;
  Result := CFNV1_32B;              // Seed
  while lLength>=4 do begin
    {$ifdef FPC_LITTLE_ENDIAN} { Intel + some others }
    Result:=Result xor SwapEndian(PDWord(lPointer)^);
    {$else FPC_LITTLE_ENDIAN}  { ARM BIG_ENDIAN + some others : ~bk : not tested  }
    Result:=Result xor PDWord(lPointer)^;
    {$endif FPC_LITTLE_ENDIAN}  { ARM BIG_ENDIAN + some others : ~bk : not tested }
    if result=0 then                // ~bk Try to reseed sticky state
      result:=CFNV1_32BS * lLength; // Use remaining count of unprocessed bytes
    Result := Result * CFNV1_32P;
    inc(lPointer, SizeOf(DWord));
    dec(lLength,SizeOf(DWord));
  end;
  if lLength>0 then begin
    lDWord:=0;
    lpByte:=@lDWord;
    while lLength>0 do begin        // ~bk pointer bug corrected 2014.03.15
      PByte(lpByte)^:=PByte(lPointer)^;
      inc(lPointer);
      inc(lpByte);
      dec(lLength);
    end;
    {$ifdef FPC_LITTLE_ENDIAN} { Intel + some others }
    Result:=Result xor SwapEndian(lDWord);
    {$else FPC_LITTLE_ENDIAN}  { ARM BIG_ENDIAN + some others }
    Result:=Result xor lDWord;
    {$endif FPC_LITTLE_ENDIAN}  { ARM BIG_ENDIAN + some others : ~bk : not tested }
    if result=0 then                // #bk Try to reseed sticky state
      result:=CFNV1_32BS * aLength; // Use total count of bytes in the buffer
    {$ifdef FPC_LITTLE_ENDIAN} { Intel + some others }
    Result:=SwapEndian(Result * CFNV1_32P);
    {$else FPC_LITTLE_ENDIAN}  { ARM BIG_ENDIAN + some others : ~bk : not tested  }
    Result := Result * CFNV1_32P;
    {$endif FPC_LITTLE_ENDIAN}  { ARM BIG_ENDIAN + some others : ~bk : not tested }
  end;
end;

{ version 1 }
function PackHash32To8(aDWord: DWord): byte;
var
  lpByte:PByte;
begin
  lpByte:=PByte(@aDword);
  Result:=lpByte^;            inc(lpByte);
  Result:=Result xor lpByte^; inc(lpByte);
  Result:=Result xor lpByte^; inc(lpByte);
  Result:=Result xor lpByte^;
end;

end.

